import aiobotocore
import gcsfs
import s3fs
from fsspec.core import logger
from fsspec.utils import setup_logging
from pyiceberg.io import PY_IO_IMPL
from pyiceberg.table import StaticTable

from universql.lake.fsspec_util import MonitoredSimpleCacheFileSystem


def s3(cache_storage: str, profile: str = "default"):
    session = aiobotocore.session.AioSession(profile=profile)
    s3_file_system = s3fs.S3FileSystem(session=session)
    caching_fs = MonitoredSimpleCacheFileSystem(
        fs=s3_file_system,
        cache_storage=cache_storage,
    )
    return caching_fs


def gcs(cache_storage, project=None, token=None):
    setup_logging(logger=logger, level="ERROR")
    gcs_file_system = gcsfs.GCSFileSystem(project=project, token=token)
    caching_fs = MonitoredSimpleCacheFileSystem(
        fs=gcs_file_system,
        cache_storage=cache_storage,
    )
    return caching_fs


CACHE_DIRECTORY_KEY = "universql.cache_directory"


def iceberg(data):
    from pyiceberg.io.fsspec import FsspecFileIO
    io = FsspecFileIO(data)
    get_fs = io.get_fs
    io.get_fs = lambda name: MonitoredSimpleCacheFileSystem(
        fs=get_fs(name),
        cache_storage=data.get(CACHE_DIRECTORY_KEY),
    )
    return io


def get_iceberg_table_from_data_lake(metadata_file_path: str, cache_directory):
    from_metadata = StaticTable.from_metadata(metadata_file_path, {
        PY_IO_IMPL: "universql.lake.cloud.iceberg",
        CACHE_DIRECTORY_KEY: cache_directory,
    })
    return from_metadata


import base64
import logging
import socket
import sys
import tempfile
import typing
from pathlib import Path

import click
import psutil
import requests
import uvicorn
from requests import RequestException

from universql.util import LOCALHOST_UNIVERSQL_COM_BYTES, Catalog, sizeof_fmt, SNOWFLAKE_HOST

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(name)s %(levelname)-6s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')
logger = logging.getLogger("🏠")

DEFAULTS = {
    "max_memory": sizeof_fmt(psutil.virtual_memory().total * 0.8),
    "max_cache_size": sizeof_fmt(psutil.disk_usage("./").free * 0.8)
}


@click.group()
@click.version_option(version="0.1")
def cli():
    pass


LOCALHOSTCOMPUTING_COM = "localhostcomputing.com"


@cli.command(
    epilog='[BETA] Check out docs at https://github.com/buremba/universql and let me know if you have any cool use-case on Github!')
@click.option('--account',
              help='The account to use. Supports both Snowflake and Polaris (ex: rt21601.europe-west2.gcp)',
              prompt='Account ID', envvar='SNOWFLAKE_ACCOUNT')
@click.option('--port', help='Port for Snowflake proxy server (default: 8084)', default=8084, type=int)
@click.option('--host', help='Host for Snowflake proxy server (default: localhostcomputing.com)',
              default=LOCALHOSTCOMPUTING_COM,
              envvar='SERVER_HOST',
              type=str)
@click.option('--catalog', type=click.Choice([e.value for e in Catalog]),
              help='Type of the Snowflake account. Automatically detected if not provided.')
@click.option('--aws-profile', help='AWS profile to access S3 (default: `default`)', type=str)
@click.option('--gcp-project',
              help='GCP project to access GCS and apply quota. (to see how to setup auth for GCP and use different accounts, visit https://cloud.google.com/docs/authentication/application-default-credentials)',
              type=str)
# @click.option('--azure-tenant', help='Azure account to access Blob Storage (ex: default az credentials)', type=str)
@click.option('--ssl_keyfile',
              help='SSL keyfile for the proxy server, optional. Use it if you don\'t want to use localhostcomputing.com',
              type=str)
@click.option('--ssl_certfile', help='SSL certfile for the proxy server, optional. ', type=str)
@click.option('--max-memory', type=str, default=DEFAULTS["max_memory"],
              help='DuckDB Max memory to use for the server (default: 80% of total memory)',
              envvar='MAX_MEMORY', )
@click.option('--cache-directory', help=f'Data lake cache directory (default: {Path.home() / ".universql" / "cache"})',
              default=Path.home() / ".universql" / "cache",
              envvar='CACHE_DIRECTORY',
              type=str)
@click.option('--max-cache-size', type=str, default=DEFAULTS["max_cache_size"],
              help='DuckDB maximum cache used in local disk (default: 80% of total available disk)',
              envvar='CACHE_PERCENTAGE', )
@click.option('--database-path', type=click.Path(exists=False, writable=True), default=":memory:",
              help='For persistent storage, provide a path to the DuckDB database file (default: :memory:)',
              envvar='DATABASE_PATH', )
def snowflake(host, port, ssl_keyfile, ssl_certfile, account, catalog, **kwargs):
    context__params = click.get_current_context().params
    auto_catalog_mode = catalog is None
    if auto_catalog_mode:
        try:
            polaris_server_check = requests.get(
                f"https://{account}.snowflakecomputing.com/polaris/api/catalog/v1/oauth/tokens")
            is_polaris = polaris_server_check.status_code == 405
        except RequestException as e:
            error_message = (
                f"Unable to find Snowflake account (https://{account}.snowflakecomputing.com), make sure if you have access to the Snowflake account. (maybe need VPN access?) \n"
                f"You can set `--catalog` property to avoid this error. \n {str(e.args)}")
            logger.error(error_message)
            sys.exit(1)

        context__params["catalog"] = Catalog.POLARIS.value if is_polaris else Catalog.SNOWFLAKE.value

    adjective = "apparently" if auto_catalog_mode else ""
    logger.info(f"UniverSQL is starting reverse proxy for {account}.snowflakecomputing.com, "
                f"it's {adjective} a {context__params['catalog']} server.")

    if host == LOCALHOSTCOMPUTING_COM:
        data = socket.gethostbyname_ex(LOCALHOSTCOMPUTING_COM)
        logger.info(f"Using the SSL keyfile and certfile for localhostcomputing.com. DNS resolves to {data}")
        if "127.0.0.1" not in data[2]:
            logger.error(
                "The DNS setting for localhostcomputing.com doesn't point to localhost, refusing to start. Please update UniverSQL.")
            sys.exit(1)

    # Don't use SSL inside Snowflake container
    if SNOWFLAKE_HOST is None or ssl_certfile is not None:
        with tempfile.NamedTemporaryFile(suffix='cert.pem', delete=True) as cert_file:
            cert_file.write(base64.b64decode(LOCALHOST_UNIVERSQL_COM_BYTES['cert']))
            cert_file.flush()
            with tempfile.NamedTemporaryFile(suffix='key.pem', delete=True) as key_file:
                key_file.write(base64.b64decode(LOCALHOST_UNIVERSQL_COM_BYTES['key']))
                key_file.flush()

                uvicorn.run("universql.protocol.snowflake:app",
                            host=host, port=port,
                            ssl_keyfile=ssl_keyfile or key_file.name,
                            ssl_certfile=ssl_certfile or cert_file.name,
                            reload=False,
                            use_colors=True)
    else:
        uvicorn.run("universql.protocol.snowflake:app",
                    host=host, port=port,
                    reload=False,
                    use_colors=True)


class EndpointFilter(logging.Filter):
    def __init__(
            self,
            path: str,
            *args: typing.Any,
            **kwargs: typing.Any,
    ):
        super().__init__(*args, **kwargs)
        self._path = path

    def filter(self, record: logging.LogRecord) -> bool:
        return record.getMessage().find(self._path) == -1


uvicorn_logger = logging.getLogger("uvicorn.access")
uvicorn_logger.addFilter(EndpointFilter(path="/telemetry/send"))
uvicorn_logger.addFilter(EndpointFilter(path="/queries/v1/query-request"))
uvicorn_logger.addFilter(EndpointFilter(path="/session"))

if __name__ == '__main__':
    cli(prog_name="universql")



import json
import logging
import typing
from typing import List
from uuid import uuid4

import pyarrow as pa
import snowflake.connector
import sqlglot
from pyarrow import ArrowInvalid
from snowflake.connector import NotSupportedError, DatabaseError
from snowflake.connector.constants import FIELD_TYPES, FieldType
from snowflake.connector.cursor import SnowflakeCursor
from sqlglot.expressions import Literal, Var, Property, IcebergProperty, Properties, ColumnDef, DataType, \
    Schema, TransientProperty, TemporaryProperty, Select, Column, Alias, Anonymous, parse_identifier, Subquery

from universql.warehouse import ICatalog, Executor, Locations, IcebergTable, CreateRelation
from universql.util import SNOWFLAKE_HOST, QueryError
from universql.protocol.utils import get_field_for_snowflake

MAX_LIMIT = 10000

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("❄️")


class SnowflakeExecutor(Executor):

    def __init__(self, query_id: str, cursor: SnowflakeCursor, cant_use_warehouse: False):
        self.query_id = query_id
        self.cursor = cursor
        self.cant_use_warehouse = cant_use_warehouse

    def supports(self, ast: sqlglot.exp.Expression) -> bool:
        return True

    def default_create_table_as_iceberg(self, expression: sqlglot.exp.Expression):
        if isinstance(expression, sqlglot.exp.Create):
            if expression.kind == 'TABLE':
                properties = expression.args.get('properties') or Properties()
                is_transient = TransientProperty() in properties.expressions
                is_temp = TemporaryProperty() in properties.expressions
                is_iceberg = IcebergProperty() in properties.expressions
                if is_transient or len(properties.expressions) == 0:
                    properties__set = Properties()
                    properties__set.set('expressions', [IcebergProperty(),
                                                        Property(this=Var(this='EXTERNAL_VOLUME'),
                                                                 value=Literal.string("iceberg_jinjat")),
                                                        Property(this=Var(this='CATALOG'),
                                                                 value=Literal.string("snowflake")),
                                                        Property(this=Var(this='BASE_LOCATION'),
                                                                 value=Literal.string(
                                                                     "iceberg_table_test")), ])

                    metadata_query = expression.expression.sql(dialect="snowflake")
                    try:
                        self.cursor.describe(metadata_query)
                    except Exception as e:
                        logger.error(f"Unable fetching schema for metadata query {e.args} \n" + metadata_query)
                        return expression
                    columns = [(column.name, FIELD_TYPES[column.type_code]) for column in
                               self.cursor.description]
                    unsupported_columns = [(column[0], column[1].name) for column in columns if column[1].name not in (
                        'BOOLEAN', 'TIME', 'BINARY', 'TIMESTAMP_TZ', 'TIMESTAMP_NTZ', 'TIMESTAMP_LTZ', 'TIMESTAMP',
                        'DATE', 'FIXED',
                        'TEXT', 'REAL')]
                    if len(unsupported_columns) > 0:
                        logger.error(
                            f"Unsupported columns {unsupported_columns} in {expression.expression.sql(dialect='snowflake')}")
                        return expression

                    column_definitions = [ColumnDef(
                        this=sqlglot.exp.parse_identifier(column[0]),
                        kind=DataType.build(self._convert_snowflake_to_iceberg_type(column[1]), dialect="snowflake"))
                        for
                        column in
                        columns]
                    schema = Schema()
                    schema.set('this', expression.this)
                    schema.set('expressions', column_definitions)
                    expression.set('this', schema)
                    select = Select().from_(Subquery(this=expression.expression))
                    for column in columns:
                        col_ast = Column(this=parse_identifier(column[0]))
                        if column[1].name in ('ARRAY', 'OBJECT'):
                            alias = Alias(this=Anonymous(this="to_variant", expressions=[col_ast]),
                                          alias=parse_identifier(column[0]))
                            select = select.select(alias)
                        else:
                            select = select.select(col_ast)

                    expression.set('expression', select)
                    expression.set('properties', properties__set)
        return expression

    def _convert_snowflake_to_iceberg_type(self, snowflake_type: FieldType) -> str:
        if snowflake_type.name == 'TIMESTAMP_LTZ':
            return 'TIMESTAMP'
        if snowflake_type.name == 'VARIANT':
            # No support for semi-structured data. Maybe we should try OBJECT([SCHEMA])?
            return 'TEXT'
        if snowflake_type.name == 'ARRAY':
            # Relies on TO_VARIANT transformation
            return 'TEXT'
        if snowflake_type.name == 'OBJECT':
            # The schema is not available
            return 'TEXT'
        return snowflake_type.name

    def execute_raw(self, compiled_sql: str) -> None:
        run_on_warehouse = not self.cant_use_warehouse
        try:
            emoji = "☁️(user cloud services)" if not run_on_warehouse else "💰(used warehouse)"
            logger.info(f"[{self.query_id}] Running on Snowflake.. {emoji} \n {compiled_sql}")
            self.cursor.execute(compiled_sql)
        except DatabaseError as e:
            message = f"{e.sfqid}: {e.msg} \n{compiled_sql}"
            raise QueryError(message, e.sqlstate)

    def execute(self, ast: sqlglot.exp.Expression,
                locations: typing.Dict[sqlglot.exp.Table, str]) \
            -> typing.Optional[typing.Dict[sqlglot.exp.Table, str]]:
        compiled_sql = ast.transform(self.default_create_table_as_iceberg).sql(dialect="snowflake", pretty=True)
        self.execute_raw(compiled_sql)
        return None

    def get_query_log(self, total_duration) -> str:
        return "Run on Snowflake"

    def close(self):
        self.cursor.close()

    @staticmethod
    def _get_ref(table_information) -> IcebergTable:
        location = table_information.get('metadataLocation')
        return IcebergTable(location)

    def get_as_table(self) -> pa.Table:
        try:
            arrow_all = self.cursor.fetch_arrow_all(force_return_table=True)
            for idx, column in enumerate(self.cursor._description):
                (field, value) = get_field_for_snowflake(column, arrow_all[idx])
                arrow_all = arrow_all.set_column(idx, field, value)
            return arrow_all
        # return from snowflake is not using arrow
        except NotSupportedError:
            row = self.cursor.fetchone()
            values = [[] for _ in range(len(self.cursor._description))]

            while row is not None:
                for idx, column in enumerate(row):
                    values[idx].append(column)
                row = self.cursor.fetchone()

            fields = []
            for idx, column in enumerate(self.cursor._description):
                (field, _) = get_field_for_snowflake(column)
                fields.append(field)
            schema = pa.schema(fields)

            result_data = pa.Table.from_arrays([pa.array(value) for value in values], names=schema.names)

            for idx, column in enumerate(self.cursor._description):
                (field, value) = get_field_for_snowflake(column, result_data[idx])
                try:
                    result_data = result_data.set_column(idx, field, value)
                except ArrowInvalid as e:
                    # TODO: find a better approach (maybe casting?)
                    if any(value is not None for value in values):
                        result_data = result_data.set_column(idx, field, pa.nulls(len(result_data), field.type))
                    else:
                        raise QueryError(f"Unable to transform response: {e}")

            return result_data


class SnowflakeCatalog(ICatalog):

    def __init__(self, context: dict, query_id: str, credentials: dict, compute):
        super().__init__(context, query_id, credentials, compute)
        if context.get('account') is not None:
            credentials["account"] = context.get('account')
        if SNOWFLAKE_HOST is not None:
            credentials["host"] = SNOWFLAKE_HOST

        self.databases = {}
        credentials["warehouse"] = compute.get('warehouse', str(uuid4()))
        try:
            self.cursor = snowflake.connector.connect(**credentials).cursor()
        except DatabaseError as e:
            raise QueryError(e.msg, e.sqlstate)

    def register_locations(self, tables: Locations):
        prefix = "gs://my-iceberg-data/custom-events/"

        queries = []
        for location, table in tables.items():
            if isinstance(location, IcebergTable):
                queries.append(f"""CREATE OR REPLACE ICEBERG TABLE {table.sql(dialect="snowflake")}
                         EXTERNAL_VOLUME='iceberg_jinjat'
                         CATALOG='icebergCatalogInt'
                         METADATA_FILE_PATH='{location.location[len(prefix):]}';""")
            elif isinstance(location, CreateRelation):
                prop_sf = [property.sql(dialect="snowflake") for property in location.properties]
                queries.append(
                    f"""CREATE OR REPLACE {' '.join(prop_sf)} {location.kind} {table.sql(dialect="snowflake")}
                                         AS {location.query.sql(dialect="snowflake")};""")
            else:
                raise QueryError("Unable to register Iceberg tables in Snowflake", "500")
        self.cursor.execute('\n'.join(queries))

    def executor(self) -> Executor:
        return SnowflakeExecutor(self.query_id, self.cursor,
                                 cant_use_warehouse=self.compute.get('warehouse') is not None)

    def get_table_paths(self, tables: List[sqlglot.exp.Table]) -> Locations:
        if len(tables) == 0:
            return {}
        sqls = ["SYSTEM$GET_ICEBERG_TABLE_INFORMATION(%s)" for _ in tables]
        values = [table.sql(comments=False, dialect="snowflake") for table in tables]
        final_query = f"SELECT {(', '.join(sqls))}"
        try:
            self.cursor.execute(final_query, values)
            result = self.cursor.fetchall()
            return {table: SnowflakeExecutor._get_ref(json.loads(result[0][idx])) for idx, table in
                    enumerate(tables)}
        except DatabaseError as e:
            err_message = f"Unable to find location of Iceberg tables. See: https://github.com/buremba/universql#cant-query-native-snowflake-tables. Cause: \n {e.msg}"
            raise QueryError(err_message, e.sqlstate)

    # def find_table_location(self, database: str, schema: str, table_name: str, lazy_check: bool = True) -> str:
    #     table_location = self.databases.get(database, {}).get(schema, {}).get(table_name)
    #     if table_location is None:
    #         if lazy_check:
    #             self.load_database_schema(database, schema)
    #             return self.find_table_location(database, schema, table_name, lazy_check=False)
    #         else:
    #             raise Exception(f"Table {table_name} not found in {database}.{schema}")
    #     return table_location
    # def load_external_volumes_for_tables(self, tables: pd.DataFrame) -> pd.DataFrame:
    #     volumes = tables["external_volume_name"].unique()
    #
    #     volume_mapping = {}
    #     for volume in volumes:
    #         volume_location = pd.read_sql("DESC EXTERNAL VOLUME identifier(%s)", self.connection, params=[volume])
    #         active_storage = duckdb.sql("""select property_value from volume_location
    #                     where parent_property = 'STORAGE_LOCATIONS' and property = 'ACTIVE'
    #                    """).fetchall()[0][0]
    #         all_properties = duckdb.execute("""select property_value from volume_location
    #                 where parent_property = 'STORAGE_LOCATIONS' and property like 'STORAGE_LOCATION_%'""").fetchall()
    #         for properties in all_properties:
    #             loads = json.loads(properties[0])
    #             if loads.get('NAME') == active_storage:
    #                 volume_mapping[volume] = loads
    #                 break
    #     return volume_mapping

    # def load_database_schema(self, database: str, schema: str):
    #     tables = self.load_iceberg_tables(database, schema)
    #     external_volumes = self.load_external_volumes_for_tables(tables)
    #
    #     tables["external_location"] = tables.apply(
    #         lambda x: (external_volumes[x["external_volume_name"]].get('STORAGE_BASE_URL')
    #                    + x["base_location"]), axis=1)
    #     if database not in self.databases:
    #         self.databases[database] = {}
    #
    #     self.databases[database][schema] = dict(zip(tables.name, tables.external_location))

    # def load_iceberg_tables(self, database: str, schema: str, after: Optional[str] = None) -> pd.DataFrame:
    #     query = "SHOW ICEBERG TABLES IN SCHEMA IDENTIFIER(%s) LIMIT %s", [database + '.' + schema, MAX_LIMIT]
    #     if after is not None:
    #         query[0] += " AFTER %s"
    #         query[1].append(after)
    #     tables = pd.read_sql(query[0], self.connection, params=query[1])
    #     if len(tables.index) >= MAX_LIMIT:
    #         after = tables.iloc[-1, :]["name"]
    #         return tables + self.load_iceberg_tables(database, schema, after=after)
    #     else:
    #         return tables
